from Bio import SeqIO
from Bio.SeqUtils import GC


class fastaFile:
    def __init__(self, fasta):
        """Class representing the fasta file

        Args:
            fasta (str): Path to fasta file
        """
        self.fasta = {}
        for record in SeqIO.parse(fasta, "fasta"):
            self.fasta[record.id] = record

    def compute_gc(self):
        """Compute GC content

        Args:
            reference (str): name of the reference sequence

        Returns:
            dict: {reference_id : gc_content}
        """
        self.gc = {}
        for ref in self.fasta:
            self.gc[ref] = GC(self.fasta[ref].seq) / 100
        return self.gc



#!/usr/bin/env python3

import click
from pydamage.main import pydamage_analyze
from pydamage.citation import get_citation
from pydamage.filter import apply_filter
from pydamage import __version__
from collections import OrderedDict


class NaturalOrderGroup(click.Group):
    def __init__(self, name=None, commands=None, **attrs):
        super(NaturalOrderGroup, self).__init__(name=name, commands=None, **attrs)
        if commands is None:
            commands = OrderedDict()
        elif not isinstance(commands, OrderedDict):
            commands = OrderedDict(commands)
        self.commands = commands

    def list_commands(self, ctx):
        return self.commands.keys()


@click.group(cls=NaturalOrderGroup)
@click.version_option(__version__)
@click.pass_context
@click.option(
    "-o",
    "--outdir",
    type=click.Path(writable=True, dir_okay=True),
    default="pydamage_results",
    show_default=True,
    help="Output directory",
)
def cli(ctx, outdir):
    """\b
    PyDamage: Damage parameter estimation for ancient DNA
    Author: Maxime Borry
    Contact: <borry[at]shh.mpg.de>
    Homepage & Documentation: github.com/maxibor/pydamage

    """

    ctx.ensure_object(dict)

    ctx.obj["outdir"] = outdir
    pass


@cli.command()
@click.pass_context
@click.argument("bam", type=click.Path(exists=True))
@click.option(
    "-w",
    "--wlen",
    default=20,
    type=int,
    show_default=True,
    help="Window length (in bp) for damage modeling",
)
@click.option(
    "-p",
    "--process",
    default=2,
    type=int,
    show_default=True,
    help="Number of processes for parallel computing",
)
@click.option(
    "-s", "--show_al", is_flag=True, help="Display alignments representations"
)
@click.option("-pl", "--plot", is_flag=True, help="Write damage fitting plots to disk")
@click.option("-vv", "--verbose", is_flag=True, help="Verbose mode")
@click.option(
    "-f", "--force", is_flag=True, help="Force overwriting of results directory"
)
@click.option(
    "-g",
    "--group",
    is_flag=True,
    help="Use entire BAM file as single reference for analyis "
    "(ignore reference headers)",
)
def analyze(ctx, no_args_is_help=True, **kwargs):
    """\b
    Run the PyDamage analysis

    BAM: path to BAM/SAM/CRAM alignment file. MD tags need to be set.
    """
    pydamage_analyze(**kwargs, **ctx.obj)


@cli.command()
@click.pass_context
@click.argument("csv", type=click.Path(exists=True))
@click.option(
    "-t",
    "--threshold",
    default=0.5,
    type=float,
    show_default=True,
    help="Predicted accuracy filtering threshold. Set to 0 for finding threshold with kneed method",
)
def filter(ctx, no_args_is_help=True, **kwargs):
    """\b
    Filter PyDamage results on predicted accuracy and qvalue thresholds.

    CSV: path to PyDamage result file
    """

    apply_filter(**kwargs, **ctx.obj)


@cli.command()
def cite():
    """Get pydamage citation in bibtex format"""
    get_citation()


if __name__ == "__main__":
    cli()



#!/usr/bin/env python3
import pysam
from pydamage import utils
import multiprocessing
from functools import partial
from pydamage import damage
from pydamage.plot import damageplot
from pydamage.exceptions import PyDamageWarning
from pydamage.accuracy_model import prepare_data, load_model, fit_model
import sys
from tqdm import tqdm
import warnings
from pydamage import __version__


# def pydamage_analyze(
#     bam,
#     wlen=30,
#     show_al=False,
#     process=1,
#     outdir="",
#     plot=False,
#     verbose=False,
#     force=False,
#     group=False,
# ):

#     if group:
#         pydamage_analyze_group(
#             bam, wlen, show_al, process, outdir, plot, verbose, force
#         )

#     else:
#         pydamage_analyze_multi(
#             bam, wlen, show_al, process, outdir, plot, verbose, force
#         )


def pydamage_analyze(
    bam,
    wlen=30,
    show_al=False,
    process=1,
    outdir="",
    plot=False,
    verbose=False,
    force=False,
    group=False,
):
    """Runs the pydamage analysis for each reference separately

    Args:
        bam(str): Path to alignment (sam/bam/cram) file
        wlen(int): window length
        show_al(bool): print alignments representations
        process(int):  Number of  processes for parellel computing
        outdir(str): Path to output directory
        verbose(bool): verbose mode
        force(bool): force overwriting of results directory
    Returns:
        pd.DataFrame: pandas DataFrame containg pydamage results

    """
    if verbose:
        print(f"Pydamage version {__version__}\n")
    utils.makedir(outdir, force=force)

    mode = utils.check_extension(bam)
    alf = pysam.AlignmentFile(bam, mode)

    if not alf.has_index():
        print(
            f"BAM file {bam} has no index. Sort BAM file and provide index "
            "before running pydamage."
        )
        sys.exit(1)

    refs = list(alf.references)

    if len(refs) == 0:
        print(f"No aligned sequences in {bam}")
        return []

    proc = min(len(refs), process)

    ##########################
    # Simple loop for debugging
    ##########################
    # filt_res = []
    # for ref in refs:
    #     res = damage.test_damage(
    #         bam=bam,
    #         ref=ref,
    #         wlen=wlen,
    #         show_al=show_al,
    #         mode=mode,
    #         process=process,
    #         verbose=verbose,
    #     )
    #     if res:
    #         filt_res.append(res)
    #     break
    ##########################
    ##########################

    test_damage_partial = partial(
        damage.test_damage,
        bam=bam,
        mode=mode,
        wlen=wlen,
        show_al=show_al,
        process=process,
        verbose=verbose,
    )
    print("Estimating and testing Damage")
    if group:
        filt_res = [
            damage.test_damage(
                ref=None,
                bam=bam,
                mode=mode,
                wlen=wlen,
                show_al=show_al,
                process=process,
                verbose=verbose,
            )
        ]
    else:
        with multiprocessing.Pool(proc) as p:
            res = list(tqdm(p.imap(test_damage_partial, refs), total=len(refs)))
        filt_res = [i for i in res if i]

    print(f"{len(filt_res)} contig(s) successfully analyzed by Pydamage")
    if len(filt_res) == 0:
        warnings.warn(
            "No alignments were found, check your alignment file", PyDamageWarning
        )

    if plot and len(filt_res) > 0:
        print("\nGenerating Pydamage plots")
        plotdir = f"{outdir}/plots"
        utils.makedir(plotdir, confirm=False)

        plot_partial = partial(damageplot, outdir=plotdir, wlen=wlen)
        with multiprocessing.Pool(proc) as p:
            list(tqdm(p.imap(plot_partial, filt_res), total=len(filt_res)))
    df_pydamage = utils.pandas_processing(res_dict=filt_res, wlen=wlen)

    acc_model = load_model()
    prep_df_glm = prepare_data(df_pydamage)
    df_glm = fit_model(prep_df_glm, acc_model)

    df = df_glm.merge(df_pydamage, left_index=True, right_index=True)

    utils.df_to_csv(df, outdir)
    return df



import warnings


class Error(Exception):
    """Base class for exceptions in this module."""

    pass


class AlignmentFileError(Error):
    """Exception raised for errors in validating the dataset agains the standards.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message


class PyDamageWarning(UserWarning):
    pass



from kneed import KneeLocator
from numpy import arange
from pydamage.utils import df_to_csv
from pandas import read_csv
import os


def define_threshold(pydam_df, min_knee=0.5, alpha=0.05):
    """Find kneedle point in PyDamage results

    Finding the kneedle point to get the optimal
    tradeoff between FP and FN, for the predicted
    accurary threshold

    Args:
        pydam_df (pandas df): pydamage results
        min_knee (float, optional): Min predicted_accuracy threshold.
        alpha(float, optional): Alpha q-value threshold
    """
    thresholds = [i.round(2) for i in arange(min_knee, 1, 0.01)]
    nb_contigs = list()
    nb_contigs = []
    for i in thresholds:
        nb_contigs.append(
            pydam_df.query(f"predicted_accuracy >= {i} & qvalue <= {alpha}").shape[0]
        )
    kneedle = KneeLocator(
        thresholds,
        nb_contigs,
        S=1.0,
        curve="convex",
        direction="decreasing",
        online=True,
    )
    print(thresholds)
    print(nb_contigs)
    return kneedle.knee


def filter_pydamage_results(pydam_df, acc_thresh, alpha=0.05):
    """Filter pydamage results on pred_accuracy and qvalue

    Args:
        pydam_df (pandas df): pydamage results
        acc_thresh (float): predicted accuracy threshold
        alpha (float, optional): Alpha q-value threshold. Defaults to 0.05.
    """

    return pydam_df.query(f"predicted_accuracy >= {acc_thresh} & qvalue <= {alpha}")


def apply_filter(csv, threshold, outdir, alpha=0.05):
    """Apply pydamage filtering

    Args:
        csv (str): path to pydamage result file
        threshold(float): Treshold value. 0 is for finding threshold with kneed method
        outdir (str): Path to output directory
        alpha (float, optional): Alpha q-value threshold. Defaults to 0.05.
    """

    df = read_csv(csv)
    outfile = "pydamage_filtered_results.csv"
    if threshold == 0:
        threshold = define_threshold(df)
        print(f"Optimal prediction accuracy threshold found to be: {threshold}")
    filt_df = filter_pydamage_results(df, acc_thresh=threshold)
    print(
        f"Filtering PyDamage results with qvalue <= {alpha} and predicted_accuracy >= {threshold}"
    )
    if not os.path.exists(outdir):
        os.mkdir(outdir)
    df_to_csv(filt_df, outdir, outfile)
    print(f"Filtered PyDamage results written to {outdir}/{outfile}")
    return filt_df



from scipy.stats import norm
from numpy import std, sqrt, log


def vuong_closeness(LA, LB, N, pdiff):
    """Performs Vuong's closeness test

    Args:
        LA (np.array): pointwise likelihood of model A
        LB (np.array): pointwise likelihood of model B
        N (int): Number of observations
        pdiff (int): Difference in number of model paramaters (A - B)
    Returns:
        float: z-score
        float: pvalue
    """

    LR = LA.sum() - LB.sum() - (pdiff / 2) * log(N)
    omega = std(LA - LB)
    Z = LR / (omega * sqrt(N))
    pval = 1 - norm.cdf(Z)
    return (Z, pval)



#!/usr/bin/env python3

import numpy as np
from pydamage.optim import optim
from scipy.stats import binom
from pydamage.utils import sort_dict_by_keys, RMSE, create_damage_dict
from pydamage.likelihood_ratio import LR


def fit_models(
    ref,
    model_A,
    model_B,
    damage,
    mut_count,
    conserved_count,
    verbose,
):
    """Performs model fitting and runs Likelihood ratio test

    Args:
        ref (str): name of referene in alignment file
        model_A (pydamage.models): Pydamage H1 Damage model
        model_B (pydamage.models): Pydamage H0 Null model
        damage (np.array of float): Amount of damage at each position where CtoT or GtoA (if reversed) transitions were observed
        mut_count(np.array of int): Number of CtoT and GtoA per position
        conserved_count(np.array of int): Number of conserved C and G per position
        verbose (bool): verbose mode
    """
    #################
    # MODEL FITTING #
    #################

    # Only fitting model to interval [0,wlen]
    xdata = np.arange(damage.size)

    res = {}
    optim_A, stdev_A = optim(
        function=model_A.fit,  # damage model
        parameters=model_A.kwds,
        xdata=xdata,
        ydata=damage,
        bounds=model_A.bounds,
    )
    if optim_A["pmax"] < optim_A["pmin"]:  # making sure that fitting makes sense
        optim_A["pmax"] = optim_A["pmin"]

    optim_B, stdev_B = optim(
        function=model_B.fit,  # null model
        parameters=model_B.kwds,
        xdata=xdata,
        ydata=damage,
        bounds=model_B.bounds,
    )

    ##########################
    # LIKELIHOOD CALCULATION #
    ##########################

    # position of sites where C in reference
    c_sites = xdata
    # counts of conserved positions at each site
    c2c_count_per_site = conserved_count
    # counts of CtoT and GtoA mutations at each site
    binom_k = mut_count
    # all C per site
    binom_n = c2c_count_per_site + binom_k

    # Likelihood for model A - Damage Model
    binom_p_damage = model_A.fit(c_sites, **optim_A)
    LA = binom.logpmf(k=binom_k, n=binom_n, p=binom_p_damage)

    # Likelihood for model B - Null Model
    binom_p_null = model_B.fit(c_sites, **optim_B)
    LB = binom.logpmf(k=binom_k, n=binom_n, p=binom_p_null)

    # Difference of number of paramters between model A and model B
    pdiff = len(model_A.kwds) - len(model_B.kwds)

    ################
    # LR TEST #
    ################

    LR_lambda, pval = LR(L0=LB, L1=LA, df=pdiff)

    res.update(optim_A)
    res.update(stdev_A)
    res.update(optim_B)
    res.update(stdev_B)
    res.update({"pvalue": pval})
    res.update(
        {
            "model_params": list(optim_A.values())
            + list(optim_B.values())
            + list(stdev_A.values())
            + list(stdev_B.values())
        }
    )
    res["residuals"] = damage - model_A.fit(x=xdata, **optim_A)
    res["RMSE"] = RMSE(res["residuals"])
    return res



#!/usr/bin/env python3

import pysam
from pydamage.parse_damage import damage_al
from pydamage.model_fit import fit_models
from pydamage import models
import numpy as np


class al_to_damage:
    def __init__(self, reference, al_handle, wlen):
        """Constructor of the class

        Args:
            reference (string): a reference from the indexed bam file
            al_handle(pysam.AlignmentFile)
            wlen (int): window length

        """
        self.alignments = al_handle.fetch(reference)
        self.reference = reference
        self.wlen = wlen
        # self.alignments = al_file

    # def __repr__(self):
    #     return(f"Reference {reference}")

    def get_damage(self, show_al):
        """Compute CtoT substitutions

        Args:
            show_al(bool): print alignments representations
        Returns:
            list: C - positions from 5' of C in reads
            list: CT- positions from 5' of CtoT transitions in reads
            list: G - positions from 3' of G in reverse reads
            list: GA- positions from 3' of GtoA transitions in reverse reads
            list: damage_bases - positions of C2T and G2A bases (if reversed)
            list: C_G_bases - positions of C2C and G2G bases (if reversed)
        """
        self.C = []
        self.CT = []
        self.damage_bases = []
        self.C_G_bases = []
        self.no_mut = []
        for al in self.alignments:
            if al.is_unmapped is False:
                all_damage = damage_al(
                    reference=al.get_reference_sequence(),
                    read_name=al.query_name,
                    is_reverse=al.is_reverse,
                    ref_name=self.reference,
                    query=al.query_sequence,
                    cigartuple=al.cigartuples,
                    wlen=self.wlen,
                    show_al=show_al,
                )
                self.C += all_damage["C"]
                self.CT += all_damage["CT"]
                self.damage_bases += all_damage["CT"]
                self.C_G_bases += all_damage["C"]
                self.no_mut += all_damage["no_mut"]

    def compute_damage(self):
        """Computes the amount of damage for statistical modelling"""

        # All C in reference
        C_pos, C_counts = np.unique(
            np.sort(
                self.C,
            ),
            return_counts=True,
        )
        C_dict = dict(zip(C_pos, C_counts))

        # CtoT transitions
        CT_pos, CT_counts = np.unique(np.sort(self.CT), return_counts=True)
        CT_dict = dict(zip(CT_pos, CT_counts))

        # All transitions
        damage_bases_pos, damage_bases_counts = np.unique(
            np.sort(self.damage_bases), return_counts=True
        )
        damage_bases_dict = dict(zip(damage_bases_pos, damage_bases_counts))

        # All C and G in reference
        C_G_bases_pos, C_G_bases_counts = np.unique(
            np.sort(self.C_G_bases), return_counts=True
        )
        C_G_bases_dict = dict(zip(C_G_bases_pos, C_G_bases_counts))

        # All conserved C and G
        no_mut_pos, no_mut_counts = np.unique(np.sort(self.no_mut), return_counts=True)
        no_mut_dict = dict(zip(no_mut_pos, no_mut_counts))

        CT_damage_amount = []
        GA_damage_amount = []
        damage_amount = []

        # Checking for non covered positions
        for i in range(self.wlen):
            if i not in CT_dict:
                CT_dict[i] = 0
            if i not in damage_bases_dict:
                damage_bases_dict[i] = 0
            if i not in no_mut_dict:
                no_mut_dict[i] = 0
            if i not in C_dict:
                CT_damage_amount.append(0)
            else:
                CT_damage_amount.append(CT_dict[i] / C_dict[i])
            if i not in C_G_bases_dict:
                damage_amount.append(0)
            else:
                damage_amount.append(damage_bases_dict[i] / C_G_bases_dict[i])

        return (
            np.array(
                list(damage_bases_dict.values())
            ),  # Number of CtoT and GtoA per position
            np.array(
                list(no_mut_dict.values())
            ),  # Number of conserved C and G per position
            np.array(CT_damage_amount),
            np.array(GA_damage_amount),
            np.array(damage_amount),
        )


def avg_coverage_contig(pysam_cov):
    """Computes average coverage of a reference

    Args:
        pysam_cov (np.array): Four dimensional array of coverage for each base
    Returns:
        float: mean coverage of reference
    """
    A = np.array(pysam_cov[0], dtype=int)
    C = np.array(pysam_cov[1], dtype=int)
    G = np.array(pysam_cov[2], dtype=int)
    T = np.array(pysam_cov[3], dtype=int)
    cov_all_bases = A + C + G + T
    cov = np.mean(cov_all_bases)
    return cov


def check_model_fit(model_dict, wlen, verbose):
    """Check if model fitting makes sense

    Args:
        model_dict (dict): Dictionary containing LR test results
        wlen (int): window length
        verbose (bool): Run in verbose mode
    Returns:
        bool or model_dict(dict): False if test fails, model_dict otherwise
    """
    # Check that no fitted parameters or stdev are infinite
    if np.isinf(np.array(model_dict["model_params"])).any():
        if verbose:
            print(f"Unreliable model fit for {model_dict['reference']}")
        return False

    # Check that all the first wlen bases are covered
    # if len(model_dict["base_cov"]) < wlen:
    #     return False
    # if np.any(np.array(model_dict["base_cov"][:wlen]) == 0):
    #     if verbose:
    #         print(
    #             f"Could not reliably fit a model to {model_dict['reference']}"
    #             "because of too few reads aligned"
    #         )
    #     return False
    return model_dict


def test_damage(ref, bam, mode, wlen, show_al, process, verbose):
    """Prepare data and run LRtest to test for damage

    Args:
        ref (str): name of referene in alignment file
        bam (str): bam file
        mode (str): opening mode of alignment file
        wlen (int): window length
        show_al (bool): Show alignment representations
        process (int): Number of process for parallelization
        verbose (bool): Run in verbose mode
    Returns:
        dict: Dictionary containing LR test results
    """
    al_handle = pysam.AlignmentFile(bam, mode=mode, threads=process)
    try:
        if ref is None:
            all_references = al_handle.references
            cov = np.mean(
                [
                    avg_coverage_contig(al_handle.count_coverage(contig=ref))
                    for ref in all_references
                ]
            )
            nb_reads_aligned = np.sum(
                [al_handle.count(contig=ref) for ref in all_references]
            )
            reflen = np.sum(
                [al_handle.get_reference_length(ref) for ref in all_references]
            )
            refname = "reference"
        else:
            cov = avg_coverage_contig(al_handle.count_coverage(contig=ref))
            nb_reads_aligned = al_handle.count(contig=ref)
            reflen = al_handle.get_reference_length(ref)
            refname = ref

        al = al_to_damage(reference=ref, al_handle=al_handle, wlen=wlen)
        al.get_damage(show_al=show_al)
        (
            mut_count,
            conserved_count,
            CT_damage,
            GA_damage,
            all_damage,
        ) = al.compute_damage()
        # if all_damage:
        model_A = models.damage_model()
        model_B = models.null_model()
        test_res = fit_models(
            ref=ref,
            model_A=model_A,
            model_B=model_B,
            damage=all_damage,
            mut_count=mut_count,
            conserved_count=conserved_count,
            verbose=verbose,
        )
        test_res["reference"] = refname
        test_res["nb_reads_aligned"] = nb_reads_aligned
        test_res["coverage"] = cov
        test_res["reflen"] = reflen

        CT_log = {}
        GA_log = {}

        for i in range(wlen):
            CT_log[f"CtoT-{i}"] = CT_damage[i]
        test_res.update(CT_log)
        test_res.update(GA_log)

        # for i in range(qlen):
        #     if i not in ydata_counts:
        #         ydata_counts[i] = np.nan
        #     if f"CtoT-{i}" not in ctot_out:
        #         ctot_out[f"CtoT-{i}"] = np.nan
        #     if f"GtoA-{i}" not in gtoa_out:
        #         gtoa_out[f"GtoA-{i}"] = np.nan

        # print(test_res)

        return check_model_fit(test_res, wlen, verbose)

    except (ValueError, RuntimeError) as e:
        if verbose:
            print(f"Model fitting for {ref} failed")
            print(f"Model fitting error: {e}")
            print(
                f"nb_reads_aligned: {nb_reads_aligned} - coverage: {cov}"
                " - reflen: {reflen}\n"
            )
        return False


# def get_damage_group(get_damage_list):
#     """Prepare data and run LR test to test for damage on reference grouped
#     as one.

#     Args:
#         get_damage_list (list): results of get_damage as list of dict
#     Returns:
#         list:[ct_data, ga_data, cc_data, c_data, g_data, all_bases]
#     """

#     al = al_to_damage(reference=ref, al_handle=al_handle)
#     (
#         CT,
#         GA,
#         damage_bases,
#         forward,
#         reverse,
#     ) = al.get_damage(wlen=wlen, show_al=show_al)

#     return [
#         CT,
#         GA,
#         damage_bases,
#         forward,
#         reverse,
#         cov,
#         nb_reads_aligned,
#         reflen,
#     ]


# def test_damage_group(
#     ct_data,
#     ga_data,
#     damage_bases,
#     forward,
#     reverse,
#     nb_reads_aligned,
#     cov,
#     reflen,
#     wlen,
#     verbose,
# ):
#     """Performs damage test

#     Args:
#         ct_data (list of int): List of positions from 5' with CtoT transitions
#         ga_data (list of int): List of positions from 3' with GtoA transitions
#         damage_bases (list of int): List of positions with damage
#         forward (list of int): List of positions from 5' where a base is aligned
#         reverse(list of int): List of positions from 3' where a base is aligned
#         nb_reads_aligned(int): number of reads aligned
#         cov(float): average coverage across all references
#         reflen(int): length of all references
#         wlen (int): window length
#         verbose(bool): Verbose
#     Returns:
#         dict: Dictionary containing LR test results
#     """
#     ref = "reference"
#     try:
#         if ct_data:
#             model_A = models.damage_model()
#             model_B = models.null_model()
#             test_res = fit_models(
#                 ref="reference",
#                 model_A=model_A,
#                 model_B=model_B,
#                 ct_data=ct_data,
#                 cc_data=cc_data,
#                 ga_data=ga_data,
#                 all_bases=all_bases,
#                 wlen=wlen,
#                 verbose=verbose,
#             )
#             test_res["reference"] = ref
#             test_res["nb_reads_aligned"] = nb_reads_aligned
#             test_res["coverage"] = cov
#             test_res["reflen"] = reflen

#             return check_model_fit(test_res, wlen, verbose)

#     except (ValueError, RuntimeError) as e:
#         if verbose:
#             print(f"Model fitting for {ref} failed")
#             print(f"Model fitting error: {e}")
#             print(
#                 f"nb_reads_aligned: {nb_reads_aligned} - coverage: {cov} "
#                 "- reflen: {reflen}\n"
#             )
#         return False



#!/usr/bin/env python3

import numpy as np


class damage_model:
    def __init__(self):
        self.kwds = ["p", "pmin", "pmax"]
        self.bounds = ((1e-15, 1e-15, 1e-15), (0.99, 0.2, 0.99))

    def __repr__(self):
        return f"""
            A modified geometric function of formula:
            y = ((((1-p)**x)*p) - xmin)/(xmax - xmin))*(pmax - pmin) + pmin
            With parameters:
            - p
            - pmin
            - pmax
            - xmin
            - xmax
            """

    def _geom_pmf(self, x, p):
        """Probability mass function of the discrete geometric distribution

        Args:
            x (int): position
            p (float): parameter of distribition
        Returns:
            float: PMF(x)
        """
        return ((1 - p) ** x) * p

    def fit(self, x, p, pmin, pmax, wlen=35):
        """Damage model function

        Args:
            x (numpy ndarray or int) data
            p (float): shape parameter
            pmin (float): min y value
            pmax (float): max y value
            wlen(int):  window length
        Returns:
            np.array: PMF(x)
        """
        vec_base_geom = np.vectorize(self._geom_pmf)

        if type(x) == int:
            x = np.array(x)

        base_geom = vec_base_geom(x, p)

        xmax = self._geom_pmf(0, p)
        xmin = self._geom_pmf(wlen - 1, p)
        scaled_geom = ((base_geom - xmin) / (xmax - xmin)) * (pmax - pmin) + pmin
        return scaled_geom


class null_model:
    def __init__(self):
        self.kwds = ("p0",)
        self.bounds = ((1e-15,), (0.2,))

    def __repr__(self):
        return f"""
            A modified uniform function of formula
            y = p0
            With parameters:
            - p0
            """

    def fit(self, x, p0):
        """Null model function

        Args:
            x (numpy array) data
            p0 (float): target y value
        Returns:
            np.array: PMF(x)
        """
        return np.array([p0] * len(x))



import numpy as np
import os
import sys
import shutil
from statsmodels.stats.multitest import multipletests
import pandas as pd
from typing import Tuple


def check_extension(filename: str) -> str:
    """Check alignment file format to give correct open mode

    Args:
        filename (str): Path to alignment file

    Returns:
        str: opening mode

    Raises:
        Exception: Extension not supported
    """
    extension = filename.split(".")[-1]
    modes = {"bam": "rb", "sam": "r", "cram": "rc"}
    try:
        return modes[extension]
    except KeyError:
        raise Exception(f"{extension} file extension not supported")


def makedir(dirpath: str, confirm: bool = True, force: bool = False):
    """Make directory if user confirms overwritting

    Args:
        dirpath (str): Path to directory
        confirm (bool, optional): Ask the user to confirm. Defaults to True.
        force (bool, optional): Always create directory. Defaults to False.
    """
    if os.path.exists(dirpath):
        if confirm and force is False:
            print(
                f"Result directory, {dirpath}, already exists, it will be overwritten"
            )
            if input("Do You Want To Continue? (y|n) ").lower() != "y":
                sys.exit()
        shutil.rmtree(dirpath)

    os.makedirs(dirpath)


def pandas_processing(res_dict: dict, wlen: int) -> pd.core.frame.DataFrame:
    """Performs Pandas processing of Pydamage results

    Args:
        res_dict (dict): Result dictionary of LR test
        wlen(int): window size
    """
    df = pd.DataFrame(res_dict)
    if len(res_dict) == 0:
        return df
    qvalues = pd.Series(
        multipletests(df["pvalue"].dropna(), method="fdr_bh")[1],
        index=df["pvalue"].dropna().index,
        name="qvalue",
    )
    df = df.merge(qvalues, left_index=True, right_index=True, how="outer")
    df = df[
        [
            "p0",
            "p0_stdev",
            "p",
            "p_stdev",
            "pmin",
            "pmin_stdev",
            "pmax",
            "pmax_stdev",
            "pvalue",
            "qvalue",
            "RMSE",
            "reference",
            "nb_reads_aligned",
            "coverage",
            "reflen",
        ]
        + [f"CtoT-{i}" for i in range(wlen)]
    ]
    df.rename(
        columns={
            "p0": "null_model_p0",
            "p0_stdev": "null_model_p0_stdev",
            "p": "damage_model_p",
            "p_stdev": "damage_model_p_stdev",
            "pmin": "damage_model_pmin",
            "pmin_stdev": "damage_model_pmin_stdev",
            "pmax": "damage_model_pmax",
            "pmax_stdev": "damage_model_pmax_stdev",
        },
        inplace=True,
    )
    df.sort_values(by=["qvalue"], inplace=True)
    df.set_index("reference", inplace=True)
    df.dropna(axis=1, how="all", inplace=True)
    df = df.round(3)
    return df


def pandas_group_processing(res_dict: dict) -> pd.core.frame.DataFrame:
    """Performs Pandas processing of Pydamage grouped reference results

    Args:
        res_dict (dict): Result dictionary of LR test
    """
    filt_dict = {}
    filt_dict["reference"] = str(res_dict["reference"])
    filt_dict["null_model_p0"] = float(res_dict["p0"])
    filt_dict["null_model_p0_stdev"] = float(res_dict["p0_stdev"])
    filt_dict["damage_model_p"] = float(res_dict["p"])
    filt_dict["damage_model_p_stdev"] = float(res_dict["p_stdev"])
    filt_dict["damage_model_pmin"] = float(res_dict["pmin"])
    filt_dict["damage_model_pmin_stdev"] = float(res_dict["pmin_stdev"])
    filt_dict["damage_model_pmax"] = float(res_dict["pmax"])
    filt_dict["damage_model_pmax_stdev"] = float(res_dict["pmax_stdev"])
    filt_dict["pvalue"] = float(res_dict["pvalue"])
    filt_dict["RMSE"] = float(res_dict["RMSE"])
    filt_dict["nb_reads_aligned"] = int(res_dict["nb_reads_aligned"])
    filt_dict["coverage"] = float(res_dict["coverage"])
    filt_dict["reflen"] = int(res_dict["reflen"])

    for i in list(res_dict.keys()):
        if str(i).startswith("CtoT"):
            filt_dict[i] = float(res_dict[i])
        if str(i).startswith("GtoA"):
            filt_dict[i] = float(res_dict[i])

    df = (
        pd.Series(filt_dict)
        .to_frame(name="reference")
        .transpose()
        .set_index("reference")
    )
    return df


def df_to_csv(
    df: pd.core.frame.DataFrame, outdir: str, outfile: str = "pydamage_results.csv"
):
    """Write Pydamage results to disk

    Args:
        df(pandas DataFrame): Pydamage results DataFrame
        outdir (str): Path to output directory
    """
    df = df.round(3)
    if not outdir:
        outdir = "."
    df.to_csv(f"{outdir}/{outfile}")


def sort_dict_by_keys(adict: dict) -> dict:
    """Sort dictonary by keys

    Args:
        adict (dict): dictorary to sort
    Returns:
        dict: dictionary sorted on keys
    """
    res = {}
    for k in sorted(adict.keys()):
        res[k] = adict[k]
    return res


def RMSE(residuals: np.ndarray) -> float:
    """Computes Root Mean Square Error

    Args:
        residuals (np.array(float)): Array of residuals
    Returns:
        float: RMSE
    """
    return np.sqrt(np.mean(residuals ** 2))


def create_damage_dict(
    damage_data: list, non_damage_data: list, wlen: int
) -> Tuple[dict, dict]:
    """Creates C bases positions dictionnary

    For C->T transitions
    For C in reference and in query

    Args:
        damage_data (list of int): List of positions where CtoT transitions were observed
        non_damage_data (list of int): List of positions where C in ref and query
    Returns:
        dict{int:int}: {position: number of CtoT transitions at position}
        dict{int:int}: {position: number of C match at position}
    """
    damage_pos, damage_counts = np.unique(np.sort(damage_data), return_counts=True)
    non_damage_pos, non_damage_counts = np.unique(
        np.sort(non_damage_data), return_counts=True
    )
    damage = dict(zip(damage_pos, damage_counts))
    non_damage = dict(zip(non_damage_pos, non_damage_counts))
    damage_dict = {}
    non_damage_dict = {}

    for i in range(wlen):
        if i in non_damage or i in damage:
            if i in damage:
                damage_dict[i] = damage[i]
            else:
                damage_dict[i] = 0
            if i in non_damage:
                non_damage_dict[i] = non_damage[i]
            else:
                non_damage_dict[i] = 0

    return (damage_dict, non_damage_dict)



import numpy as np
from scipy.stats import chi2


def LR(L0, L1, df):
    """Performs likelihood ratio test

    Args:
        L0 (np array): Log likelihood of null model
        L1 (np array): Log likelihood of alternative model
        df (int): chi2 degrees of freedom: number of constrained params
    Returns:
        float: LR test statistic lambda
        float: pvalue
    """

    LR_lambda = -2 * (L0.sum() - L1.sum())
    pval = 1 - chi2.cdf(LR_lambda, df)
    return (LR_lambda, pval)



from matplotlib import use
import matplotlib.pyplot as plt
import matplotlib.ticker as mticker
from pydamage.models import damage_model, null_model
from statsmodels.nonparametric.smoothers_lowess import lowess
from pydamage import utils
import numpy as np
from os import makedirs
from scipy.stats import probplot

use("Agg")


def damageplot(damage_dict, wlen, outdir):
    """Draw pydamage plots

    Args:
        damage_dict(dict): pydamage result dictionary
        wlen (int): window length
        qlen(int): query length
        outdir(str): Pydamage result directory
    """
    x = np.array(range(wlen))
    # qlen = np.array(range(damage_dict["qlen"]))
    c2t = np.array([damage_dict[f"CtoT-{i}"] for i in x])
    y = c2t
    # g2a = np.array([damage_dict[f"GtoA-{i}"] for i in x])
    p0 = damage_dict["p0"]
    p0_stdev = damage_dict["p0_stdev"]
    p = damage_dict["p"]
    pmin = damage_dict["pmin"]
    pmin_stdev = damage_dict["pmin_stdev"]
    pmax = damage_dict["pmax"]
    pmax_stdev = damage_dict["pmax_stdev"]
    contig = damage_dict["reference"]
    pvalue = damage_dict["pvalue"]
    coverage = damage_dict["coverage"]
    residuals = damage_dict["residuals"]
    rmse = damage_dict["RMSE"]
    plotdir = outdir

    if pvalue < 0.001:
        rpval = "<0.001"
    else:
        rpval = f"={round(pvalue,3)}"

    m_null = null_model()
    p0_low = max(m_null.bounds[0][0], p0 - 2 * p0_stdev)
    p0_high = min(m_null.bounds[1][0], p0 + 2 * p0_stdev)
    y_unif = m_null.fit(x, p0)
    y_unif_low = np.maximum(np.zeros(y_unif.shape[0]), m_null.fit(x, p0_low))
    y_unif_high = np.minimum(np.ones(y_unif.shape[0]), m_null.fit(x, p0_high))

    geom = damage_model()
    geom_pmin_low = max(geom.bounds[0][1], pmin - 2 * pmin_stdev)
    geom_pmin_high = min(geom.bounds[1][1], pmin + 2 * pmin_stdev)
    geom_pmax_low = max(geom.bounds[0][2], pmax - 2 * pmax_stdev)
    geom_pmax_high = min(geom.bounds[1][2], pmax + 2 * pmax_stdev)

    y_geom = geom.fit(x, p, pmin, pmax)
    y_geom_low = np.maximum(
        np.zeros(y_geom.shape[0]), geom.fit(x, p, geom_pmin_low, geom_pmax_low)
    )
    y_geom_high = np.minimum(
        np.ones(y_geom.shape[0]), geom.fit(x, p, geom_pmin_high, geom_pmax_high)
    )

    plt.xticks(rotation=45, fontsize=8)

    ## Plot null model
    fig, ax = plt.subplots()

    ax.plot(
        x, y_unif, linewidth=2.5, color="DarkOliveGreen", alpha=0.8, label="Null model"
    )

    ax.fill_between(
        x,
        y_unif_low,
        y_unif_high,
        color="DarkOliveGreen",
        alpha=0.1,
        label="Null Model CI (2 sigma)",
    )

    ## Plot damage model
    ax.plot(x, y_geom, linewidth=2.5, color="#D7880F", alpha=0.8, label="Damage model")

    ax.fill_between(
        x,
        y_geom_low,
        y_geom_high,
        color="#D7880F",
        alpha=0.1,
        label="Damage Model CI (2 sigma)",
    )

    ## Plot real data

    ax.plot(x, c2t, color="#bd0d45", alpha=0.2, label="C to T transitions")

    ax.set_xlabel("Base from 5'", fontsize=10)
    ax.set_ylabel("Substitution frequency", fontsize=10)
    ax.set_xticks(np.arange(x[0], x[-1], 5))
    ax.set_xticklabels(ax.get_xticks(), rotation=45, fontsize=6)
    ax.set_title(f"coverage: {round(coverage,2)} - pvalue{rpval}", fontsize=8)
    ax.legend(fontsize=8)
    # ax.set_title(f"coverage: {round(coverage,2)} | pvalue{rpval}", fontsize=8)

    left, bottom, width, height = [0.65, 0.4, 0.2, 0.2]
    ax2 = fig.add_axes([left, bottom, width, height])

    fitted = x
    smoothed = lowess(residuals, fitted)
    ax2.scatter(fitted, residuals, marker=".", color="black")
    ax2.plot(smoothed[:, 0], smoothed[:, 1], color="r")
    ax2.plot([min(fitted), max(fitted)], [0, 0], color="k", linestyle=":", alpha=0.3)
    ax2.set_ylabel("Residuals", fontsize=6)
    ax2.set_xlabel("Fitted Values", fontsize=6)
    ax2.set_title(f"Residuals vs. Fitted\nRMSE={round(rmse, 3)}", fontsize=6)
    ax2.set_xticks(np.arange(fitted[0], fitted[-1], 5))
    ax2.set_xticklabels([int(i) for i in ax2.get_xticks()], fontsize=6, rotation=45)
    ax2.set_yticks(ax2.get_yticks())
    ax2.set_yticklabels([round(i, 3) for i in ax2.get_yticks()], fontsize=6)

    fig.suptitle(contig, fontsize=12, y=0.95)

    fig.savefig(f"{plotdir}/{contig}.png", dpi=200)



def get_citation():
    BIB = """
@article{Borry2021_pydamage,
    author = {Borry, Maxime and Huebner, Alexander and Rohrlach, Adam B and Warinner, Christina G},
    doi = {10.1101/2021.03.24.436838},
    elocation-id = {2021.03.24.436838},
    eprint = {https://www.biorxiv.org/content/early/2021/03/24/2021.03.24.436838.full.pdf},
    journal = {bioRxiv},
    publisher = {Cold Spring Harbor Laboratory},
    title = {PyDamage: automated ancient damage identification and estimation for contigs in ancient DNA de novo assembly},
    url = {https://www.biorxiv.org/content/early/2021/03/24/2021.03.24.436838},
    year = {2021}
}
"""
    print(BIB)


#!/usr/bin/env python3


def damage_al(
    reference,
    read_name,
    is_reverse,
    ref_name,
    query,
    cigartuple,
    wlen,
    show_al,
):
    """Compute CtoT mutations for a single alignment

    Args:
        reference (str): reference sequence
        read_name(str): name of read
        is_reverse(bool): alignment if reverse
        ref_name(str): name of reference
        query (string): query sequence
        cigartuple (tuple): cigar tuple (pysam)
        wlen (int): window length
        print_al (bool): print alignment
    Returns:
        dict : {'C':  [ C pos from 5'],
                'CT': [ CtoT pos from 5'],
                'G':  [ G pos from 3'],
                'GA': [GtoA pos from 3'],
                'no_mut': [Positions where C , and G (if reversed) are conserved]}
    """
    r_pos = 0
    q_pos = 0
    r_string = ""
    q_string = ""
    res = ""
    base_trans_counts = {"C": [], "CT": [], "no_mut": []}
    for c in cigartuple:
        # [M, =, X] - alignment match (can be a sequence match or mismatch)
        if c[0] in [0, 7, 8]:
            r_string += reference[r_pos : r_pos + c[1]]
            q_string += query[q_pos : q_pos + c[1]]
            r_pos += c[1]
            q_pos += c[1]
        # I, S - Insertion of soft clipping
        if c[0] in [1, 4]:
            r_string += " " * c[1]
            q_string += query[q_pos : q_pos + c[1]]
            r_pos = r_pos
            q_pos += c[1]
        # D - Deletion in query: add space in query string
        if c[0] in [2, 3]:
            r_string += reference[r_pos : r_pos + c[1]]
            q_string += " " * c[1]
            r_pos += c[1]
            q_pos = q_pos

    read_len = len(r_string)
    for i in range(min(read_len, wlen)):
        r_char = r_string[i].upper()
        q_char = q_string[i].upper()
        # base_trans_counts[q_char].append(i)
        if is_reverse is False:
            if r_char == "C":
                base_trans_counts["C"].append(i)
                if q_char == "T":
                    base_trans_counts["CT"].append(i)
                elif q_char == "C":
                    base_trans_counts["no_mut"].append(i)
    if show_al:
        orient = {False: ["5'", "3'"], True: ["3'", "5'"]}
        res += "R   " + r_string + "\t" + ref_name + "\n" + orient[is_reverse][0] + "  "
        for i in range(0, min(len(r_string), len(q_string))):
            if r_string[i] == q_string[i]:
                res += "|"
            else:
                res += " "
        res += orient[is_reverse][1] + "\nQ   " + q_string + "\t" + read_name + "\n "
        print(res)

    return base_trans_counts



import pkg_resources
from pypmml import Model


def load_model():
    """Returns the pmml model"""
    model_path = pkg_resources.resource_stream(
        __name__, "models/pydamage_glm_model.pmml"
    )
    model = Model.load(model_path)
    return model


def prepare_data(pd_df):
    """Prepare pydamage result data for accuracy modelling

    Args:
        pd_df (pandas DataFrame):pydamage df result
    """
    pd_df = pd_df[["coverage", "reflen", "damage_model_pmax"]].rename(
        columns={
            "damage_model_pmax": "damage",
            "reflen": "contiglength",
            "coverage": "actual_cov",
        }
    )
    pd_df = pd_df.astype({"actual_cov": float, "contiglength": int, "damage": float})
    return pd_df


def fit_model(df, model):
    """Fit GLM model to data

    Args:
        df (pandas DataFrame): prepared pydamage results
        model (pypmml model): GLM accuracy model
    """
    prediction = list(model.predict(df)["Predicted_sig"])
    df["predicted_accuracy"] = prediction
    return df["predicted_accuracy"].to_frame()



from setuptools import setup, find_packages
import codecs
import os.path


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


setup(
    name="pydamage",
    version=get_version("pydamage/__init__.py"),
    description="Damage parameter estimation for ancient DNA",
    long_description=open("README.md").read(),
    url="https://github.com/maxibor/pydamage",
    long_description_content_type="text/markdown",
    license="GNU-GPLv3",
    python_requires=">=3.6",
    install_requires=[
        "click",
        "numpy",
        "pandas",
        "pysam",
        "scipy",
        "statsmodels",
        "matplotlib",
        "tqdm",
        "biopython",
        "kneed",
        "pypmml",
    ],
    packages=find_packages(include=["pydamage"]),
    entry_points={"console_scripts": ["pydamage = pydamage.cli:cli"]},
    include_package_data=True,
    package_data={"": ["models/pydamage_glm_model.pmml"]},
)



#!/usr/bin/env python3


from scipy.optimize import curve_fit
from numpy import sqrt, diag


def optim(function, parameters, xdata, ydata, bounds, loss="huber"):
    """Find optimal parameters given data

    Args:
        function (function): function to optimize
        xdata (np array): x values
        ydata (np array): y values
        bounds (tuple of tuple): optimization bounds
                    ((par1_min, par2_min), (par1_max, par2_max))
        loss (str): loss function. See https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.least_squares.html#scipy.optimize.least_squares
    Returns:
        dict: popt_dict - 'parameter_name':'parameter_value'
        dict: perr_dict - 'parameter_name':'standard_deviation'
    """
    popt, pcov = curve_fit(function, xdata=xdata, ydata=ydata, bounds=bounds, loss=loss)
    perr = sqrt(diag(pcov))
    popt_dict = {k: v for (k, v) in zip(parameters, popt)}
    perr_dict = {f"{k}_stdev": v for (k, v) in zip(parameters, perr)}
    return (popt_dict, perr_dict)
